#include <windows.h>
#include <stdio.h>
#include <memory.h>

#include "wnaspi32.h"
#include "scsidefs.h"
#include "fpse.h"

#ifdef __GCC__

typedef struct                                  // Offset
{                                               // HX/DEC
    UINT8       SRB_Cmd;                        // 00/000 ASPI command code = SC_GETSET_TIMEOUTS
    UINT8       SRB_Status;                     // 01/001 ASPI command status byte
    UINT8       SRB_HaId;                       // 02/002 ASPI host adapter number
    UINT8       SRB_Flags;                      // 03/003 ASPI request flags
    UINT32      SRB_Hdr_Rsvd;                   // 04/004 Reserved, MUST = 0
    UINT8       SRB_Target;                     // 08/008 Target's SCSI ID
    UINT8       SRB_Lun;                        // 09/009 Target's LUN number
    UINT32      SRB_Timeout;                    // 0A/010 Timeout in half seconds
} __attribute__ ((packed))
gSRB_GetSetTimeouts, *gPSRB_GetSetTimeouts, FAR *gLPSRB_GetSetTimeouts;

#define SRB_GetSetTimeouts gSRB_GetSetTimeouts
#define PSRB_GetSetTimeouts gPSRB_GetSetTimeouts
#define LPSRB_GetSetTimeouts gLPSRB_GetSetTimeouts

#endif

typedef struct {
	UINT8 DevType;
	char  reserved[7];
	char  Vendor[24];
} Inquiery;

typedef struct {
	unsigned char reserved;
	unsigned char adr;
	unsigned char trackno;
	unsigned char reserved1;
	unsigned char addr[4];
} TOCENTRY;

typedef struct {
	unsigned char size[2];
	unsigned char first;
	unsigned char last;
	TOCENTRY track[100];
} TOC;

typedef	struct {
	UINT8 minute;
	UINT8 second;
	UINT8 frame;
} LOC;

#define	INT2BCD(a)	(((a)/10)*16+((a)%10))
#define	BCD2INT(a)	(((a)>>4)*10+((a)&0xf))

#define SECTOR_SIZE 2352

static int adapter_id = 0;
static int target_id = 0;
static int target_lun = 0;
static unsigned char readbuf[SECTOR_SIZE];

#define	MAX_ID	8

int loc2int(LOC *loc);
static void int2loc(LOC *loc,int n);
static int GetDeviceType(int adapterid,int targetid);
static void ExecReadTOC(void *buf,int size,int first);
static void ExecRead(unsigned long frame,void *buf,int bufsize);
static void ExecPlay(unsigned char *start,unsigned char *end);
static void ExecStop(void);
static void ExecModeSelect(int mode);

static HMODULE WNASPI32;

DWORD (*pGetASPI32SupportInfo)( VOID );
DWORD (*pSendASPI32Command)( LPSRB );
BOOL  (*pGetASPI32Buffer)( PASPI32BUFF );
BOOL  (*pFreeASPI32Buffer)( PASPI32BUFF );
BOOL  (*pTranslateASPI32Address)( PDWORD, PDWORD );

int cd_init(void)
{
    DWORD dwSupportInfo;
    int status, hostcount, adapterid, targetid;

    WNASPI32 = LoadLibrary( "WNASPI32" );

    if( !WNASPI32 )
    {
        printf("ASPI load fail\n");
        return -1;
    }

    pGetASPI32SupportInfo   = GetProcAddress(WNASPI32,"GetASPI32SupportInfo");
    pSendASPI32Command      = GetProcAddress(WNASPI32,"SendASPI32Command");
    pGetASPI32Buffer        = GetProcAddress(WNASPI32,"GetASPI32Buffer");
    pFreeASPI32Buffer       = GetProcAddress(WNASPI32,"FreeASPI32Buffer");
    pTranslateASPI32Address = GetProcAddress(WNASPI32,"TranslateASPI32Address");

    dwSupportInfo = pGetASPI32SupportInfo();

    status = (dwSupportInfo>>8)&255;
    hostcount = dwSupportInfo&255;

    if (hostcount ==0 || status != SS_COMP) {
        printf("ASPI init fail %x",dwSupportInfo);
        return -1;
    }
    for(adapterid=0;adapterid<hostcount;adapterid++) {
        for(targetid=0;targetid<MAX_ID;targetid++) {
            int type = GetDeviceType(adapterid,targetid);

			if (type>=0) {
				printf("%d %d %d\n",adapterid,targetid,type);
			}

			if (type == DTYPE_CDROM) {
				printf("adapter=%d target=%d\n",adapterid,targetid);
				adapter_id = adapterid;
				target_id = targetid;
				return 0;
			}
		}
	}
	/* not found cdrom */
	return -1;
}

int cd_term(void)
{
    FreeLibrary(WNASPI32);
    return 0;
}

static TOC toc;
static struct {
	unsigned char size[2];
	unsigned char first,last;
	TOCENTRY track[1];
} readout;

void dump(unsigned char *data,int len)
{
	int i;
	for(i=0;i<len;i++) {
		printf("%02x ",data[i]);
		if ((i&15)==15) printf(":%04x\n",i&~15);
	}
	printf("\n");
}

int cd_GetTN(unsigned char *result)
{
	ExecReadTOC(&readout,sizeof(readout),0xaa);
	ExecReadTOC(&toc,sizeof(toc),0);
	toc.track[toc.last-toc.first+1] = readout.track[0];
/*
	{ int i;
	for(i=0;i<=toc.last;i++) dump(&toc.track[i],8);
	}
*/
	result[0] = toc.first;
	result[1] = toc.last;

    return 0;
}

int cd_GetTD(unsigned char *result,int track)
{
	int n;
	// ExecReadTOC(&toc,sizeof(toc));
	n = track-toc.first;
	result[0] = toc.track[n].addr[1]; /* miniute */
	result[1] = toc.track[n].addr[2]; /* second */
	result[2] = toc.track[n].addr[3]; /* frame */

    return 0;
}

int cd_play(unsigned char *param)
{
	/* serch next track */
	ExecPlay(param,&readout.track[0].addr[2]);
    return 0;
}

int cd_stop(void)
{
/* serch next track */
    ExecStop();
    return 0;
}

int cd_read(unsigned char *param)
{
    ExecRead(loc2int((LOC *)param),readbuf,SECTOR_SIZE);
    return 0;
}

static	SRB_ExecSCSICmd	srb;

static int ExecSCSICmd_sub(UINT8 *cdb,int cdb_size,UINT8 *out_buff,int out_size,BYTE flag)
{
	int status;

	HANDLE hEventSRB;

	hEventSRB = CreateEvent(NULL,TRUE,FALSE,NULL);
	if (hEventSRB==NULL) {
		printf("can't create event\n");
	}

	memset(&srb,0,sizeof(SRB_ExecSCSICmd));

	srb.SRB_Cmd = SC_EXEC_SCSI_CMD;
	srb.SRB_HaId = adapter_id;
	srb.SRB_Flags = flag | SRB_EVENT_NOTIFY;
	srb.SRB_Target = target_id;
//	srb.SRB_Lun = target_lun;
	srb.SRB_SenseLen = SENSE_LEN;
	srb.SRB_PostProc = (LPVOID)hEventSRB;

	srb.SRB_BufPointer = out_buff;
	srb.SRB_BufLen = out_size;
	srb.SRB_CDBLen = cdb_size;
	memcpy(srb.CDBByte,cdb,cdb_size);

	status = pSendASPI32Command(&srb);
	printf("status:%x \n",status);
	if (status == SS_PENDING) {
		WaitForSingleObject(hEventSRB,INFINITE);
		printf("wait\n");
	}
	if (srb.SRB_Status != SS_COMP) {
		printf("%x:error %x\n",cdb[0],srb.SRB_Status);
	}
	CloseHandle(hEventSRB);
    return 0;
}

static int ExecSCSICmd(UINT8 *cdb,int cdb_size,UINT8 *out_buff,int out_size)
{
	return ExecSCSICmd_sub(cdb,cdb_size,out_buff,out_size,SRB_DIR_IN);
}

static int loc2int(LOC *loc)
{
	return ( BCD2INT(loc->minute)*60 + BCD2INT(loc->second) -2 )*75 + BCD2INT(loc->frame);
}

static void int2loc(LOC *loc,int n)
{
	int minute = n/75+2;
	loc->frame = INT2BCD(n%75);
	loc->minute = INT2BCD(minute/60);
	loc->second = INT2BCD(minute%60);
}

static int GetDeviceType(int adapterid,int targetid)
{
	SRB_GDEVBlock srbGDEVBlock;
	memset(&srbGDEVBlock,0,sizeof(srbGDEVBlock));
	srbGDEVBlock.SRB_Cmd = SC_GET_DEV_TYPE;
	srbGDEVBlock.SRB_HaId = adapterid;
	srbGDEVBlock.SRB_Target = targetid;
	pSendASPI32Command((LPSRB)&srbGDEVBlock);
	if (srbGDEVBlock.SRB_Status != SS_COMP) return -1;
	return srbGDEVBlock.SRB_DeviceType;
}

static void ExecReadTOC(void *buf,int size,int first)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = SCSI_READ_TOC;
	cdb[1] = 2; /* MSF formst */
	cdb[6] = first;
	cdb[7] = size>>8;
	cdb[8] = size&255;

	ExecSCSICmd(cdb,sizeof(cdb),buf,size);
}

static void ExecRead(unsigned long frame,void *buf,int bufsize)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = SCSI_READ10;
	cdb[2] = frame>>24;
	cdb[3] = frame>>16;
	cdb[4] = frame>>8;
	cdb[5] = frame;
	cdb[7] = 0;
	cdb[8] = 1; /* read 1 frames */

	ExecSCSICmd(cdb,sizeof(cdb),buf,bufsize);
}

static void ExecPlay(unsigned char *start,unsigned char *end)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = SCSI_PLAYAUDMSF;
	cdb[3] = start[0];
	cdb[4] = start[1];
	cdb[5] = start[2];
	cdb[6] = end[0];
	cdb[7] = end[1];
	cdb[8] = end[2];

	ExecSCSICmd(cdb,sizeof(cdb),NULL,0);
}

static void ExecStop(void)
{
	UINT8 cdb[10];

	memset(cdb,0,sizeof(cdb));

	cdb[0] = 0x4B; /* Pause command */
	cdb[8] = 0; /* 0:pause 1:resume */

	ExecSCSICmd(cdb,sizeof(cdb),NULL,0);
}

static void ExecModeSelect(int mode)
{
	UINT8 cdb[6],param[12+4];

	memset(cdb,0,sizeof(cdb));
	memset(param,0,sizeof(param));

	cdb[0] = SCSI_MODE_SEL6;
	cdb[1] = 0x10; /* SCSI-2 Page format */
	cdb[4] = 12; //sizeof(param);

	param[3] = 8; /* size of block descpritor */
	param[4] = mode; /* dencity code 0:default 1:2048 2:2336 3:2340 4:audio */
	if (mode==0 || mode==1) {
		param[10] = 2048>>8;
		param[11] = 2048;
	} else {
		param[10] = SECTOR_SIZE>>8;
		param[11] = SECTOR_SIZE;
	}

	ExecSCSICmd_sub(cdb,sizeof(cdb),param,12 /*sizeof(param)*/,SRB_DIR_OUT);
}

#if 0
int main(int argc,char **argv)
{
	unsigned char result[8];
	unsigned char param[8];
	int i;

	cd_init();
	cd_GetTN(result);
	param[0] = 0;
	param[1] = 2;
	param[2] = 0;

	memset(readbuf,1,sizeof(readbuf));

	ExecModeSelect(0);

	for(i=0;i<256;i++) {
		ExecRead(i);
		printf("frame %d:\n",i);
		dump(readbuf,SECTOR_SIZE);
	}
	cd_term();
}
#endif
