/*
	cdrom.c	by	BERO
	info from psx.rules.org / doomed@c64.org
*/

#include <stdio.h>
#include <string.h>

#include "fpse.h"

#define	RAMADDR(addr) ((void*)&ram[(addr)&0xffffff])

enum {
	CdlSync,CdlNop,CdlSetloc,CdlPlay,
	CdlForward,CdlBackword,CdlReadN,CdlStandby,
	CdlStop,CdlPause,CdlReset,CdlMute,
	CdlDemute,CdlSetfilter,CdlSetmode,CdlGetparam,
	CdlGetlocL,CdlGetlocP,Cdl18,CdlGetTN,
	CdlGetTD,CdlSeekL,CdlSeekP,CdlReadS=27
};

enum {
	ModeSpeed     = 0x80,        /* 0: normal speed  1: double speed */
	ModeRT        = 0x40,        /* 0: ADPCM off     1: ADPCM on     */
	ModeSize1     = 0x20,        /* 0: 2048 byte     1: 2340 byte    */
	ModeSize0     = 0x10,        /* 0: -             1: 2328 byte    */
	ModeSF        = 0x08,        /* 0: Channel off   1: Channel on   */
	ModeRept      = 0x04,        /* 0: Report off    1: Report on    */
	ModeAP        = 0x02,        /* 0: AutoPause off 1: AutoPause on */
	ModeDA        = 0x01         /* 0: CD-DA off     1: CD-DA on     */
};

enum {
	StatPlay      = 0x80,        /* playing CD-DA */
	StatSeek      = 0x40,        /* seeking */
	StatRead      = 0x20,        /* reading data frames */
	StatShellOpen = 0x10,        /* once shell open */
	StatSeekError = 0x04,        /* seek error detected */
	StatStandby   = 0x02,        /* spindle motor rotating */
	StatError     = 0x01         /* command error detected */
};

enum {
	NoIntr      = 0x00,        /* No interrupt */
	DataReady   = 0x01,        /* Data Ready */
	Acknowledge = 0x02,        /* Acknowledge */
	Complete    = 0x03,        /* Command Complete */
	DataEnd     = 0x04,        /* End of Data Detected */
	DiskError   = 0x05        /* Error Detected */
};

struct {
	UINT8 stat;
	UINT8 ctrl;
	int reset;

	UINT8 mode;
	UINT8 param[8];
	int paramptr;
	/* param */
	UINT8 result[8];
	int resultptr;
	int resultsize;
	int resultready;
	struct {
		UINT8 minute;
		UINT8 second;
		UINT8 frame;
	} loc;
	UINT8	file;
	UINT8	chan;
} cd;

static char readbuf[2340];


/*
	result[0]

	0x80 CD-DA Play
	0x02 working?

*/

int cd0_read(void)
{
	int ret = cd.ctrl;
	if (cd.resultsize && cd.resultready) ret|=0x20; else ret&=~0x20;
	return ret|0x08;
}

int cd1_read(void)
{
	int ret=0;
	if (cd.resultsize) {
	  ret = cd.result[cd.resultptr++];
	  if (cd.resultptr>=cd.resultsize) {
	    cd.resultptr = cd.resultsize = 0;
		cd.resultready = 0;
	  }
	}
	return ret;
}


int cd2_read(void)
{
	return 0;
}

int cd3_read(void)
{
	return cd.stat;
}


#define	RESULT(n)	cd.result[0] = n; cd.resultsize = 1

#define	INT2BCD(a)	((a/10)*16+(a%10))
#define	BCD2INT(a)	((a>>4)*10+(a&0xf))

void cd0_write(int data)
{
/*	write 0??
	write read & 3?
*/
	cd.ctrl = data;
	if (data==0) {
		cd.paramptr = 0;
		cd.resultready = 0;
	}
	if (data&1) cd.resultready = 1;
	if (data==1) cd.reset = 1;
}

void cd2_write(int data)
{
	if (cd.reset==2 && data==7) {
		cd.paramptr = 0;
		cd.resultready = 1;
		cd.reset = 0;
		cd.stat = 0;
		cd.ctrl = 0;
		return;
	}
	cd.reset = 0;
	if (cd.paramptr<8)
		cd.param[cd.paramptr++] = data;
}

void cd3_write(int data)
{
//	cd.stat = data;
	if (data==7 && cd.reset==1) cd.reset=2;
	else cd.reset=0;
	if (cd.resultready && data==7) {
		cd.resultsize = cd.resultptr = 0;
		cd.stat = 0;
	}
}

void cd1_write(int data)
{
	static char *cmdname[]= {
	"CdlSync","CdlNop","CdlSetloc","CdlPlay",
	"CdlForward","CdlBackword","CdlReadN","CdlStandby",
	"CdlStop","CdlPause","CdlReset","CdlMute",
	"CdlDemute","CdlSetfilter","CdlSetmode","CdlGetparam",
	"CdlGetlocL","CdlGetlocP","Cdl18","CdlGetTN",
	"CdlGetTD","CdlSeekL","CdlSeekP","Cdl23",
	"Cdl24","Cdl25","Cdl26","CdlReadS"
	};

	if (data<28) {
		PRINTF("%s\n",cmdname[data]);
	}

	switch(data){
	case CdlNop:
		RESULT(2);
		break;

	case CdlSetloc:
		cd.loc.minute = cd.param[0];
		cd.loc.second = cd.param[1];
		cd.loc.frame = cd.param[2];
		RESULT(0);
		break;

	case CdlPlay:
		cd_play((unsigned char *)&cd.loc);
		RESULT(0x80);
		break;

	case CdlForward:
	case CdlBackword:
		RESULT(2);
		break;

	case CdlReadN:
		RESULT(2);
		break;

	case CdlStandby:
		break;

	case CdlStop:
		break;

	case CdlPause:
	case CdlReset:
	case CdlMute:
	case CdlDemute:
		RESULT(2);
		break;

	case CdlSetfilter:
		cd.file = cd.param[0];
		cd.chan = cd.param[1];
		RESULT(2);
		break;

	case CdlSetmode:
		cd.mode = cd.param[0];
		RESULT(0);
		break;

	case CdlGetparam:
		cd.result[0] = 0; /* status */
		cd.result[1] = cd.mode;
		cd.resultsize = 2;
		break;

	case CdlGetlocL:
		cd.result[0] = cd.loc.minute; /* min */
		cd.result[1] = cd.loc.second; /* sec */
		cd.result[2] = cd.loc.frame; /* frame  */
		cd.result[3] = cd.mode; /* mode */
		cd.result[4] = cd.file; /* file */
		cd.result[5] = cd.chan; /* channel */
		cd.resultsize = 6;
		break;

	case CdlGetlocP:
		cd.result[0] = 0; /* track */
		cd.result[1] = 0; /* index */
		cd.result[2] = 0; /* min  */
		cd.result[3] = 0; /* sec */
		cd.result[4] = 0; /* frame */
		cd.result[5] = 0; /* min  */
		cd.result[6] = 0; /* sec */
		cd.result[7] = 0; /* frame */
		cd.resultsize = 8;
		break;

	case CdlGetTN:
		{unsigned char result[3];
		cd_GetTN(result);
		cd.result[0] = 0; /* status */
		cd.result[1] = INT2BCD(result[1]); /* first track */
		cd.result[2] = INT2BCD(result[2]); /* total track */
		cd.resultsize = 3;
		}
		break;

	case CdlGetTD:
		{unsigned char result[8];
		cd_GetTD((unsigned char *)cd.param,(int)result);
		cd.result[0] = 0; /* status */
		cd.result[1] = INT2BCD(result[0]); /* min */
		cd.result[2] = INT2BCD(result[1]); /* sec */
		cd.resultsize = 3;
		}
		break;

	case CdlSeekL:
	case CdlSeekP:
		RESULT(2);
		break;

	case CdlReadS:
		cd_read((unsigned char *)&cd.loc);
		RESULT(2);
		break;

	default:
		printf("unknown command %d\n",data);
	}
	cd.resultptr = 0;
	cd.resultready = 1;
        cd.stat = 3;
	/* cd_interrupt happen */
        Irq_Pulse |= (1<<INT_CD);
}

void dma3_exec(UINT32 adr,UINT32 bcr,UINT32 chcr)
{
	int size;

	size = (bcr>>16)*(bcr&0xffff);
	switch(chcr) {
	case 0x01000201:
		memcpy(RAMADDR(chcr),readbuf,size);
		break;
	}
}

