#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "fpse.h"

#ifndef TRUE
#define	TRUE	1
#define	FALSE	0
#endif

unsigned char *rom,*ram,*scratchpad,*extrom,*spumem;
UINT16 *vram;
// extern int stop;

int FPSE_Flags = EMULATE_BIOS | COMPILE;
/*
char disasmflg    = FALSE;
char emulate_bios = TRUE;
char verbose      = FALSE;
char debug        = FALSE;
char in_slot      = 0;
*/

extern int breakpoint;

static char extrom_init[0x100] = {
	0xb4,0x00,0x00,0x1f,0x4c,0x69,0x63,0x65,0x6e,0x73,0x65,0x64,0x20,0x62,0x79,0x20,
	0x53,0x6f,0x6e,0x79,0x20,0x43,0x6f,0x6d,0x70,0x75,0x74,0x65,0x72,0x20,0x45,0x6e,
	0x74,0x65,0x72,0x74,0x61,0x69,0x6e,0x6d,0x65,0x6e,0x74,0x20,0x49,0x6e,0x63,0x2e,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0xb4,0x00,0x00,0x1f,0x4c,0x69,0x63,0x65,0x6e,0x73,0x65,0x64,0x20,0x62,0x79,0x20,
	0x53,0x6f,0x6e,0x79,0x20,0x43,0x6f,0x6d,0x70,0x75,0x74,0x65,0x72,0x20,0x45,0x6e,
	0x74,0x65,0x72,0x74,0x61,0x69,0x6e,0x6d,0x65,0x6e,0x74,0x20,0x49,0x6e,0x63,0x2e,
	0x00,0x00,0x00,0x00,0x08,0x00,0xe0,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
};

void memalloc(void)
{
	rom = malloc(0x80000);
	extrom = malloc(0x20000);
	memset(extrom,0xff,0x20000);
	ram = malloc(0x200000);
	scratchpad = malloc(0x1000);
//	vram = malloc(1024*512*2);
	spumem = malloc(512*1024);
}

void memfree(void)
{
	free(extrom);
	free(rom);
	free(ram);
	free(scratchpad);
//	free(vram);
	free(spumem);
}

void go(void)
{
/*
	while(!stop) {
		if (update_counter()) Interrupt();
		doInst();
	}
*/
 while(!stop) {
    if (update_counter()) Interrupt();
    doInst();
 }
}

static char helpmsg[] = 
	"fpse [-option] [file]\n"
	" file         PSX-EXE file\n"
	" -c           compiler(default)\n"
	" -i           interpriter\n"
	" -b<biosfile> bios filename\n"
	" -v           verbose\n"
	" -d           disassemble\n"
        " -e           run debugger\n"
	"run PSX-EXE with clone bios, or original bios from start\n"
	"to stop, press ctrl+C\n";

int main(int argc,char *argv[])
{
	FILE *fp;
	char *file = NULL;
	char *biosfile = NULL;
	int i;

//	breakpoint = 0x8003ace8;

	printf("PlayStation Emulator by BERO\n");
	if (argc<2) {
		printf(helpmsg);
		return -1;
	}

	for(i=1;i<argc;i++) {
		char *s = argv[i];
		if (s[0]=='-') {
			switch(s[1]) {
			case 'c': FPSE_Flags |= COMPILE;   break;
			case 'i': FPSE_Flags &= ~COMPILE;  break;
			case 'v': FPSE_Flags |= VERBOSE;   break;
			case 'd': FPSE_Flags |= DISASMFLG; break;
			case 'b': biosfile = s+2; break;
                        case 'e': FPSE_Flags |= DEBUG;     break;
			default:
				printf(helpmsg);
				return -1;
			}
		} else {
			file = s;
		}
	}

	if (hw_init() != FPSE_OK)
            {
                printf("Couldn't initialize the harware!\n");
                hw_close();
                return -1;
            }

	memalloc();
	compile_init();
	Reset();

        if (biosfile) {
            fp = fopen(biosfile,"rb");
            if (fp==NULL) {
                printf("can't load biosfile\n");
            } else {
                FPSE_Flags &= ~EMULATE_BIOS;
                fread(rom,1,0x80000,fp);
                fclose(fp);
// Run the BIOS
                i = FPSE_Flags;
		FPSE_Flags &= ~(verbose | disasmflg);
                if (!emulate_bios)
                    while(PC != 0x80030000) doInst();
                FPSE_Flags = i;
            }
        }

	if ((fp = fopen("libps.exe","rb"))!=NULL) {
		EXE_HEADER head;
		fread(&head,1,sizeof(head),fp);
		if (memcmp(head.id,"PS-X EXE",8)==0) {
			printf("%x-%x\n",(int)head.t_addr,(int)head.t_size);
			fread(&ram[head.t_addr&0x1fffff],1,head.t_size,fp);
		}
		fclose(fp);
	}

	if (file) fp = fopen(file,"rb");
        if (fp != NULL) {
		EXE_HEADER head;

		fread(&head,1,sizeof(head),fp);
		if (memcmp(head.id,"PS-X EXE",8)==0) {
			fread(&ram[head.t_addr&0x1fffff],1,head.t_size,fp);
			setpc(head.pc0);
			bios_init();
			reg.r[30] = reg.r[29] = 0x801fff00; /* defailt stack */
			reg.r[31] = 0;
		}
		fclose(fp);

	} else
        if (file != NULL)
            printf("Can't load %s\n",file);

	memcpy(extrom,extrom_init,sizeof(extrom_init));

        if (debug) run_debug();
                else
	if (compile) compile_go();
                else go();

	memfree();

        hw_close();

        return 0;
}

#if 1
int printlog(char *fmt, ...)
{
	int ret;

	va_list argp;
	va_start(argp,fmt);
	ret = vprintf(fmt,argp);
	va_end(argp);

	return ret;
}

#else
int printlog(char *fmt, ...)
{
	int ret;

	va_list argp;
	static char buf[256],prevbuf[256];
	static char *bufp=buf;
	static int count;

	va_start(argp,fmt);
	ret = vsprintf(bufp,fmt,argp);
	va_end(argp);

	bufp += strlen(bufp);
	if (bufp[-1] != '\n') {
		return ret;
	}

	if (strcmp(buf,prevbuf)==0) {
		count++;
	} else {
		if (count) { printf(" %d times\n",count+1); count=0; }
		printf("%s",buf);
		strcpy(prevbuf,buf);
	}
	bufp=buf;

	return ret;
}

#endif
