#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include "gpu.h"
#include "poly2.h"
#include "win.h"

#include "fpse.h"

/*
	GPU:
*/

#define GPUNULL 0

static unsigned char primsizeT[256] =
{
	0,1,3,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* 00 */
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* 10 */
	4,4,4,4,7,7,7,7,5,5,5,5,9,9,9,9,	/* 20 */
	6,6,6,6,9,9,9,9,8,8,8,8,12,12,12,12,/* 30 */
	3,3,3,3,0,0,0,0,5,5,5,5,6,6,6,6,	/* 40 */
	4,4,4,4,0,0,0,0,7,7,7,7,9,9,9,9,	/* 50 */
	3,3,3,3,4,4,4,4,2,2,2,2,0,0,0,0,	/* 60 */
	2,2,2,2,3,3,3,3,2,2,2,2,3,3,3,3,	/* 70 */
	4,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* 80 */
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* 90 */
	3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* a0 */
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* b0 */
	3,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* c0 */
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* d0 */
	0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,	/* e0 */
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,	/* f0 */
};

#define	SIGN(code,basebit,bit)	(int)( (code)<<(32-basebit-bit) ) >> (32-bit)

typedef struct {
	int x,y;
} POS;

typedef struct {
	int x,y,w,h;
} AREA;

struct {
	UINT32 status;
	UINT32 info[8];

	POS dispoffset;
	POS drawoffset;
	AREA screen;
	AREA disp;
	POS clip_min;
	POS clip_max;
	AREA texwindow;

	/* primtive recieve buffer */
	UINT32 primdata[16];
	int	primsize;
	int primptr;

	int retval;
} gpu;

enum {NOTRANS,MEM2VRAM,VRAM2MEM};

#define	RAMADR(addr) ((void*)&ram[addr&0xffffff])

#define	PAL	0

/*
	gpu status
    0 - 10 : draw mode (e0 prim )
    11,12: mask mode
    12-15: ??
    16 - 22: disp mode
	23: Display (1=off)
    24-25: ??
	26: GPU ready (1=ready)
	27: storeimage ready (1=ready)
	28: command ready (1=ready)
	29-30: DMA mode
    31: in interlace mode, 1:odd 0:even
*/

void gp1_write(UINT32 code)
{
	PRINTF("gpu_reg1 write %08x:",(int)code);
	switch(code>>24) {
	case 0x00:	/* All Reset */
		gpu.status = 0x14802000;
		/* com ready | GPU ready | display off 
		256x240, NTSC, 15bit, non-interlace,
		tpage = 0, dither off, can't draw to display area, mask off
		*/
		win_resize(256,240);
		PRINTF("All Reset\n");
		break;
	case 0x01:	/* Command Reset */
		gpu.primptr = gpu.primsize = 0;
		PRINTF("Command Reset\n");
		break;
	case 0x02:	/* IRQ Reset */
		PRINTF("IRQ Reset\n");
		break;
	case 0x03:	/* Display Mask */
		/* bit 1 to status bit 23 */
		gpu.status = (gpu.status & ~(1<<23)) | ((code&1)<<23);
		PRINTF("display %s\n",code&1?"off":"on");
		break;
	case 0x04:	/* DMA mode */
		/* bit 0,1 to status bit 29,30 */
		gpu.status = (gpu.status & ~(3<<29)) | ((code&3)<<29);
		{ static char *dmamode[] =
			{"not use","CPU->GPU(linear?)","CPU->GPU","GPU->CPU"};
		PRINTF("DMA mode = %s\n",dmamode[code&3]);
		}
		break;
	case 0x05:	/* disp offset */
		gpu.dispoffset.x = code & 0x3ff;
		gpu.dispoffset.y = (code>>10) & 0x3ff;
		win_offset(gpu.dispoffset.x,gpu.dispoffset.y);
		PRINTF("display offset:%d,%d\n",gpu.dispoffset.x,gpu.dispoffset.y);
		break;
	case 0x06: {	/* H start/end */
		int s,e;
		s = code&0xfff;
		e = (code>>12)&0xfff;
		gpu.screen.x = (s - 608)/10;
		gpu.screen.w = (e-s)/10;
		PRINTF("H start = %d width = %d\n",gpu.screen.x,gpu.screen.w);
		}
		break;
	case 0x07: {	/* V start/end */
		int s,e;
		s = code&0x3ff;
		e = (code>>10)&0x3ff;
		gpu.screen.y = s - PAL?19:16;
		gpu.screen.h = e-s;
		PRINTF("V start = %d height = %d\n",gpu.screen.y,gpu.screen.h);
		}
		break;
	case 0x08:	/* display mode */
		/* bit 0..5 to status bit 17..22  */
		/* bit    6 to status bit 16      */
		gpu.status = (gpu.status &~0x7f0000) | ((code&0x3f)<<17) | ((code&0x40)<<(16-6));
		{ static int dispw[] = {256,320,512,640};
		if (code&0x40) gpu.disp.w = 368;
		else gpu.disp.w = dispw[code&3];
		}
		gpu.disp.h = (code&4)?480:240;
		win_resize(gpu.disp.w,gpu.disp.h);
		PRINTF("disp size(%d,%d) %s %s %s\n",gpu.disp.w,gpu.disp.h,
			(code&0x08)?"PAL":"NTSC",
			(code&0x10)?"24bit":"16bit",
			(code&0x20)?"interlace":"non-interlace"
		);
		break;
	case 0x10:	/* GPU Info */
		switch(code&0xffffff) {
		case 3:
		case 4:
		case 5:
			gpu.retval = gpu.info[code&0xffffff];
			break;
		case 7:
			gpu.retval = 2; /* version */
			break;
		}
		break;
	default:
		printf("unknown %x\n",(int)code);
		break;
	}
}

static int px,py,x_start,y_start,x_end,y_end,inTrans;
static unsigned short *pvram;

static short clutflg[0x8000/16];
static long tpageflg;

/*
	return true if tpage changed
	and clear changed flag

	check 1page if  4bit texture
	check 2page if  8bit texture
	check 4page if 15bit texture
*/
int tpage_changed(int tpage)
{
	int changed,mask;
	static int mask0[] = {1,3,15};
	/* mode = 0:1page 1:2page 3:4page */

	mask = mask0[(tpage>>7)&3];
	tpage&=31;
	changed = tpageflg &   (mask<<tpage);
	if (changed)  tpageflg &= ~(mask<<tpage);
	return changed;
}

/*
	return true if clut changed
	and clear changed flag
*/
int clut_changed(int clut)
{
	int changed;
	changed = clutflg[clut/16] &   (1<<(clut&15));
	if (changed)  clutflg[clut/16] &= ~(1<<(clut&15));
	return changed;
}

static void vramchange(int x0,int y0,int x1,int y1)
{
	int x,y;
	for(y=y0 ;y<y1; y++) {
		for(x=x0/16; x<=(x1-1)/16; x++)
		//	clutflg[x+y*64]=1;
			clutflg[(x+y*64)/16] |= 1<<(x&15);
	}
	for(y=y0/256; y<=(y1-1)/256; y++) {
		for(x=x0/64; x<=(x1-1)/64; x++)
		//	tpageflg[x+y*16]=1;
			tpageflg |= 1<<(x+y*16);
	}
	vramchange_sub(x0,y0,x1,y1);
}

static void gpu_unknown(UINT32 *prim)
{
	PRINTF("unknown prim\n");
}

static void gpu_01(UINT32 *prim)
{
	PRINTF("Tcache Flush\n");
}

static void gpu_LoadImage(UINT16 *sp)
{
	/* LoadImage */
	PRINTF("LoadImage:(%d,%d) (%dx%d)\n",sp[2],sp[3],sp[4],sp[5]);
	inTrans = MEM2VRAM;
	px = x_start = sp[2];
	py = y_start = sp[3];
	x_end = x_start + sp[4];
	y_end = y_start + sp[5];
	pvram = &vram[py*1024];
}

static void gpu_MoveImage(UINT16 *sp)
{
	int x0,y0,x1,y1,w,h;
	x0 = sp[2]&1023;
	y0 = sp[3]& 511;
	x1 = sp[4]&1023;
	y1 = sp[5]& 511;
	w  = sp[6]&1023;
	h  = sp[7]& 511;
	PRINTF("MoveImage:(%d,%d)->(%d,%d) (%dx%d)\n",
		x0,y0,x1,y1,w,h);
	move(x0,y0,x1,y1,w,h);
	vramchange(x1,y1,x1+w,y1+h);
}

static void gpu_StoreImage(UINT16 *sp)
{
		/* StoreImage */
		PRINTF("StoreImage:(%d,%d) (%dx%d)\n",sp[2],sp[3],sp[4],sp[5]);
		inTrans = VRAM2MEM;
		px = x_start = sp[2];
		py = y_start = sp[3];
		x_end = x_start + sp[4];
		y_end = y_start + sp[5];
		pvram = &vram[py*1024];
}

static void gpu_e1(UINT32 *prim)
{
	UINT32 code = prim[0];
	/* bit 0..10 to status bit 0..10 */
	gpu.status = ( gpu.status &~0x7ff) | (code&0x7ff);
	settpage(code&0x7ff);
	PRINTF("texture page\n");
}
static void gpu_e2(UINT32 *prim)
{
	UINT32 code = prim[0];
	gpu.texwindow.w = ((code&31))*8;
	gpu.texwindow.h = ((code>>5)&31)*8;
	gpu.texwindow.x = 256-((code>>10)&31)*8;
	gpu.texwindow.y = 256-((code>>15)&31)*8;
	PRINTF("texture window\n");
}
static void gpu_e3(UINT32 *prim)
{
	UINT32 code = prim[0];
	gpu.info[3] = code&0xffffff;
	gpu.clip_min.x = code&0x3ff;
	gpu.clip_min.y = (code>>10)&0x1ff;
	setclip(gpu.clip_min.x,gpu.clip_min.y,gpu.clip_max.x,gpu.clip_max.y);
	PRINTF("clip_min (%d,%d)\n",gpu.clip_min.x,gpu.clip_min.y);
}
static void gpu_e4(UINT32 *prim)
{
	UINT32 code = prim[0];
	gpu.info[4] = code&0xffffff;
	gpu.clip_max.x = code&0x3ff;
	gpu.clip_max.y = (code>>10)&0x1ff;
	setclip(gpu.clip_min.x,gpu.clip_min.y,gpu.clip_max.x,gpu.clip_max.y);
	PRINTF("clip_max (%d,%d)\n",gpu.clip_max.x,gpu.clip_max.y);
}
static void gpu_e5(UINT32 *prim)
{
	UINT32 code = prim[0];
	gpu.info[5] = code&0xffffff;
	gpu.drawoffset.x = ((int)code<<(32-11))>>(32-11);
	gpu.drawoffset.y = ((int)code<<(32-22))>>(32-11);
	setoffset(gpu.drawoffset.x,gpu.drawoffset.y);
	PRINTF("draw offset:%d,%d\n",gpu.drawoffset.x,gpu.drawoffset.y);
}
static void gpu_e6(UINT32 *prim)
{
	UINT32 code = prim[0];
	/* bit 0,1 to status bit 11,12 */
	gpu.status = (gpu.status & ~(3<<11)) | ((code&3)<<11);
	PRINTF("maskmode:%d\n",(int)(code&3));
}

#if !GPUNULL
static void (*primfunc[256])() = {
	/* 00 */
	gpu_unknown,gpu_01,gpu_fill,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 10 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 20 */
	gpu_polyF3,gpu_polyF3,gpu_polyF3,gpu_polyF3,
	gpu_polyFT3,gpu_polyFT3,gpu_polyFT3,gpu_polyFT3,
	gpu_polyF4,gpu_polyF4,gpu_polyF4,gpu_polyF4,
	gpu_polyFT4,gpu_polyFT4,gpu_polyFT4,gpu_polyFT4,
	/* 30 */
	gpu_polyG3,gpu_polyG3,gpu_polyG3,gpu_polyG3,
	gpu_polyGT3,gpu_polyGT3,gpu_polyGT3,gpu_polyGT3,
	gpu_polyG4,gpu_polyG4,gpu_polyG4,gpu_polyG4,
	gpu_polyGT4,gpu_polyGT4,gpu_polyGT4,gpu_polyGT4,
	/* 40 */
	gpu_lineF2,gpu_lineF2,gpu_lineF2,gpu_lineF2,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_lineF3,gpu_lineF3,gpu_lineF3,gpu_lineF3,
	gpu_lineF4,gpu_lineF4,gpu_lineF4,gpu_lineF4,
	/* 50 */
	gpu_lineG2,gpu_lineG2,gpu_lineG2,gpu_lineG2,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_lineG3,gpu_lineG3,gpu_lineG3,gpu_lineG3,
	gpu_lineG4,gpu_lineG4,gpu_lineG4,gpu_lineG4,
	/* 60 */
	gpu_tile,gpu_tile,gpu_tile,gpu_tile,
	gpu_sprt,gpu_sprt,gpu_sprt,gpu_sprt,
	gpu_tile1,gpu_tile1,gpu_tile1,gpu_tile1,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 70 */
	gpu_tile8,gpu_tile8,gpu_tile8,gpu_tile8,
	gpu_sprt8,gpu_sprt8,gpu_sprt8,gpu_sprt8,
	gpu_tile16,gpu_tile16,gpu_tile16,gpu_tile16,
	gpu_sprt16,gpu_sprt16,gpu_sprt16,gpu_sprt16,
	/* 80 */
	gpu_MoveImage,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 90 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* a0 */
	gpu_LoadImage,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* b0 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* c0 */
	gpu_StoreImage,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* d0 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* e0 */
	gpu_unknown,gpu_e1,gpu_e2,gpu_e3,
	gpu_e4,gpu_e5,gpu_e6,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* f0 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
};
#else
static void (*primfunc[256])() = {
	/* 00 */
	gpu_unknown,gpu_01,gpu_fill,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 10 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 20 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 30 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 40 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 50 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 60 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 70 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 80 */
	gpu_MoveImage,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* 90 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* a0 */
	gpu_LoadImage,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* b0 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* c0 */
	gpu_StoreImage,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* d0 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* e0 */
	gpu_unknown,gpu_e1,gpu_e2,gpu_e3,
	gpu_e4,gpu_e5,gpu_e6,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	/* f0 */
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
	gpu_unknown,gpu_unknown,gpu_unknown,gpu_unknown,
};
#endif

static void do_prim(void *prim)
{
	primfunc[*(UINT32*)prim>>24](prim);
}

void gpu_update(void)
{
	gpu.status^=0x80000000;
	/* interlace odd/even */
}

UINT32 gp1_read(void)
{
	return (gpu.status | 0x1c000000) & ~0x00480000;
	// isinterheight=480 off ꂪinterlasewait̂
	/* GPU_Idle | GPU_cmd_ready | img ready */
	 //|0x800;
	/* nazo */
}

UINT32 gp0_read(void)
{
	if (inTrans==VRAM2MEM) {
		int tmp;
		tmp = pvram[px];
		if (++px>=x_end) {
			px = x_start;
			pvram += 1024;
			if (++py>=y_end) inTrans=0;
		}
		tmp |= pvram[px]<<16;
		if (++px>=x_end) {
			px = x_start;
			pvram +=1024;
			if (++py>=y_end) inTrans=0;
		}
		gpu.retval = tmp;
	}
	return gpu.retval;
}

void gp0_write(UINT32 data)
{
	PRINTF("gpu_reg0 write %08x\n",(int)data);

	if (inTrans==MEM2VRAM) {
		pvram[px]=data;
		if (++px>=x_end) {
			px = x_start;
			pvram += 1024;
			if (++py>=y_end) inTrans=0;
		}
		if (inTrans) {
			pvram[px]=data>>16;
			if (++px>=x_end) {
				px = x_start;
				pvram += 1024;
				if (++py>=y_end) inTrans=0;
			}
		}
		if (inTrans==0) {
			vramchange(x_start,y_start,x_end,y_end);
		}
		return;
	}
	if (gpu.primptr) {
		gpu.primdata[gpu.primptr++] = data;
		if (gpu.primptr == gpu.primsize) {
			gpu.primptr = gpu.primsize = 0;
			do_prim(gpu.primdata);
		}
	} else {
		int primsize=primsizeT[data>>24];
		gpu.primdata[0] = data;
		switch(primsize) {
		case 0:
			printf("unknown prim:%08x\n",(int)data);
			break;
		case 1:
			do_prim(gpu.primdata);
			break;
		default:
			gpu.primsize = primsize;
			gpu.primptr = 1;
			break;
		}
	}
}

static void dma_chain(UINT32 addr)
{
	do {
                addr &= 0x1FFFFF;
                {
		UINT32 *ptr = (UINT32*)&ram[addr];
		UINT32 data = *ptr++;
		int size = data>>24;
//		PRINTF("packet:%x,size=%d\n",(int)addr,size);
		while (size) {
			int primsize;
			do_prim(ptr);
			primsize = primsizeT[*ptr>>24];
                        if (primsize) {
                            ptr  += primsize;
                            size -= primsize;
                        } else {
                            ptr++;
                            size--;
                        }
		}
		addr = data&0xffffff;
//                PRINTF("link:%x\n",(int)addr);
                }
		if (addr==0) {PRINTF("broken\n"); /*stop=1;*/ break; }
	} while(addr != 0xffffff);
}

static void dma_mem2vram(UINT32 *ptr,int size)
{
  while(size) {
	if (inTrans) {
		while (size--) {
			UINT32 data = *ptr++;
			pvram[px] = data;
			if (++px>=x_end) {
				px = x_start;
				pvram += 1024;
				if (++py>=y_end) {
					inTrans = 0;
					break;
				}
			}
			pvram[px] = data>>16;
			if (++px>=x_end) {
				px = x_start;
				pvram += 1024;
				if (++py>=y_end) {
					inTrans = 0;
					break;
				}
			}
		}
		if (inTrans==0) {
			vramchange(x_start,y_start,x_end,y_end);
		}
	} else {
		int primsize;
		do_prim(ptr);
		primsize = primsizeT[*ptr>>24];
		ptr  += primsize;
		size -= primsize;
	}
  }
}

static void dma_vram2mem(UINT16 *ptr,int size)
{
	size*=2;
	do {
		*ptr++ = pvram[px];
		if (++px>=x_end) {
			px = x_start;
			pvram += 1024;
			if (++py>=y_end) {
				inTrans = 0;
				break;
			}
		}
	} while (--size);
}

void gpudma_exec(UINT32 adr,UINT32 bcr,UINT32 chcr)
{
	switch(chcr){
	case 0x01000200:
		PRINTF("dma vram2mem: %x %x %x\n",(int)adr,(int)bcr,(int)chcr);
		dma_vram2mem((void*)&ram[adr&0x1fffff],(bcr>>16)*(bcr&0xffff));
		break;
	case 0x01000201:
		PRINTF("dma mem2vram: %x %x %x\n",(int)adr,(int)bcr,(int)chcr);
		dma_mem2vram((void*)&ram[adr&0x1fffff],(bcr>>16)*(bcr&0xffff));
		break;
	case 0x01000401:
		PRINTF("dma chain: %x %x %x\n",(int)adr,(int)bcr,(int)chcr);
		dma_chain(adr&0x1fffff);
		break;
	default: printf("gpudma unknown %08x\n",(int)chcr);
	}
//	PRINTF("%x %x %x\n",(int)adr,(int)bcr,(int)chcr);
}

void gpu_init(void)
{
	memset(&gpu,0,sizeof(gpu));
	gpu.info[7] = 2; /* type */
	DrawInit();
}

void ScreenShot(void)
{
 FILE   *bmpfile;
 char    filename[256];
 UINT16 *vram2;
 UINT8   header[0x36];
 UINT32  size;
 UINT8   line[1024*3];
 UINT8   empty[2]={0,0};
 UINT16  color;
 int     snapshotnr = 0,i,j;
 
 size=gpu.disp.h*gpu.disp.w*3+0x38;
 
 // fill in proper values for BMP
 
 // hardcoded BMP header
 memset(header,0,0x36);
 header[0]='B';
 header[1]='M';
 header[2]=size&0xff;
 header[3]=(size>>8)&0xff;
 header[4]=(size>>16)&0xff;
 header[5]=(size>>24)&0xff;
 header[0x0a]=0x36;
 header[0x0e]=0x28;
 header[0x12]=gpu.disp.w%256;
 header[0x13]=gpu.disp.w/256;
 header[0x16]=gpu.disp.h%256;
 header[0x17]=gpu.disp.h/256;
 header[0x1a]=0x01;
 header[0x1c]=0x18;
 header[0x26]=0x12;
 header[0x27]=0x0B;
 header[0x2A]=0x12;
 header[0x2B]=0x0B;
 
 for (;;) {
   if (++snapshotnr > 9999) return;
   sprintf(filename,"snap%04d.bmp",snapshotnr);
   bmpfile=fopen(filename,"rb");
   if (bmpfile == NULL) break;
   fclose(bmpfile);
  }

 printf("snap:%s\n",filename);

 // try opening new snapshot file
 if((bmpfile=fopen(filename,"wb"))==NULL)
  {
   // could not open for writing !
   return;
  }
 
 fwrite(header,0x36,1,bmpfile);
 vram2=vram + (gpu.disp.h+gpu.disp.y-1)*1024 + gpu.disp.x;
 for(i=0;i<gpu.disp.h;i++)
  {
   for(j=0;j<gpu.disp.w;j++)
    {
     color=vram2[j];
     line[j*3+2]= (color >> 7) & 0xF1;
     line[j*3+1]= (color >> 2) & 0xF1;
     line[j*3+0]= (color << 3) & 0xF1;
    }
   fwrite(line,gpu.disp.w*3,1,bmpfile);
   vram2 -= 1024;
  }
 fwrite(empty,0x2,1,bmpfile);
 fclose(bmpfile);  
}	    
