#include <stdio.h>
#include <time.h>

#include "fpse.h"

#define	HW16(a) (*(UINT16*)&hwarea[a])
#define	HW32(a) (*(UINT32*)&hwarea[a])

#define AUTOSPEED   0
#define NOPRINTPC   0

typedef struct {
	unsigned long addr;
	char *name;
} NAM;

/*
	1f801000
	1f801004
	1f801008
	1f80100c
	1f801010
	1f80101c
	1f802041	???
*/
static NAM hwnames[] = {
	{0x1f801014,"spu_delay"},
	{0x1f801018,"dv5_delay"},
	{0x1f801020,"com_delay"},

	{0x1f801040,"sio0_data"},
	{0x1f801044,"sio0_status"},
	{0x1f80104a,"sio0_control"},
	{0x1f80104e,"sio0_baud"},

	{0x1f801050,"sio1_data"},
	{0x1f801054,"sio1_status"},
	{0x1f80105a,"sio1_control"},
	{0x1f80105e,"sio1_baud"},

	{0x1f801060,"ram_size"},

	{0x1f801070,"int_reg"},
	{0x1f801074,"int_mask"},
	
	{0x1f801080,"mdec_dma0_madr"},
	{0x1f801084,"mdec_dma0_bcr"},
	{0x1f801088,"mdec_dma0_chcr"},
	
	{0x1f801090,"mdec_dma1_madr"},
	{0x1f801094,"mdec_dma1_bcr"},
	{0x1f801098,"mdec_dma1_chcr"},
	
	{0x1f8010a0,"gpu_dma_madr"},
	{0x1f8010a4,"gpu_dma_bcr"},
	{0x1f8010a8,"gpu_dma_chcr"},
	
	{0x1f8010b0,"cd_dma_madr"},
	{0x1f8010b4,"cd_dma_bcr"},
	{0x1f8010b8,"cd_dma_chcr"},
	
	{0x1f8010c0,"spu_dma_madr"},
	{0x1f8010c4,"spu_dma_bcr"},
	{0x1f8010c8,"spu_dma_chcr"},
	
	{0x1f8010d0,"dma5_madr"},
	{0x1f8010d4,"dma5_bcr"},
	{0x1f8010d8,"dma5_chcr"},
	
	{0x1f8010e0,"dma6_madr"},
	{0x1f8010e4,"dma6_bcr"},
	{0x1f8010e8,"dma6_chcr"},
	
	{0x1f8010f0,"dma_pcr"},
	{0x1f8010f4,"dma_icr"},

	{0x1f801100,"t0_count"},
	{0x1f801104,"t0_mode"},
	{0x1f801108,"t0_target"},

	{0x1f801110,"t1_count"},
	{0x1f801114,"t1_mode"},
	{0x1f801118,"t1_target"},

	{0x1f801120,"t2_count"},
	{0x1f801124,"t2_mode"},
	{0x1f801128,"t2_target"},

	{0x1f801800,"cdrom_reg0"},
	{0x1f801801,"cdrom_reg1"},
	{0x1f801802,"cdrom_reg2"},
	{0x1f801803,"cdrom_reg3"},

	{0x1f801810,"gpu_reg0"},
	{0x1f801814,"gpu_reg1"},

	{0x1f801820,"mdec_reg0"},
	{0x1f801824,"mdec_reg1"},

	{0x1f801d80,"spu_mvol_l"},
	{0x1f801d82,"spu_mvol_r"},
	{0x1f801d84,"spu_reverb_l"},
	{0x1f801d86,"spu_reverb_r"},
	{0x1f801d88,"spu_key_on_1"},
	{0x1f801d8a,"spu_key_on_2"},
	{0x1f801d8c,"spu_key_off_1"},
	{0x1f801d8e,"spu_key_off_2"},
	{0x1f801d90,"spu_key_modefm_1"},
	{0x1f801d92,"spu_key_modefm_2"},
	{0x1f801d94,"spu_key_modenoise_2"},
	{0x1f801d96,"spu_key_modenoise_2"},
	{0x1f801d98,"spu_key_modereverb_1"},
	{0x1f801d9a,"spu_key_modereverb_2"},
	{0x1f801d9c,"spu_key_channelactive_1"},
	{0x1f801d9e,"spu_key_channelactive_2"},

	{0x1f801da6,"spu_sbaddr"},
	{0x1f801da8,"spu_data"},
	{0x1f801daa,"spu_reg0"},
	{0x1f801dac,"spu_reg1"},
	{0x1f801dae,"spu_status"},

	{0x1f801db0,"spu_cdvol_l"},
	{0x1f801db2,"spu_cdvol_r"},
	{0x1f801db4,"spu_extvol_l"},
	{0x1f801db6,"spu_extvol_r"},

	{0x1f801dc0,"spu_reverbconfig"},
	{0x1f801dfc,"spu_factor_l"},
	{0x1f801dfe,"spu_factor_r"},

	{0x1f802030,"int_2000"},
	{0x1f802040,"dip_switches"},

	/* extension */
	{0x1f000000,"parrom_status"},
	{0x1f000001,"parrom_type"},
	{0x1f005555,"parrom_reg0"},
	{0x1f002aaa,"parrom_reg1"},
	{0x1f060000,"par_in"}, /* byte */
	{0x1f060008,"par_out"}, /* byte */
	{0x1f020010,"par_status"}, /* bit 0: 1= busy */
	{0x1f020018,"par_switch"}, /* bit 0: 1= on */

	{0,""}
};

static char *sregname[] = {
	"vol_l","vol_r","freq","sbadr","ar","adsr2","adsrvol","repeat"
};

char* hwname(unsigned long addr)
{
	NAM *p;
	static char buf[16];
	for(p=hwnames; p->addr && p->addr!=addr; p++) ;
	if (p->addr) return p->name;
	if (addr>=0x1f801c00 && addr<0x1f801c00+16*24) {
		sprintf(buf,"voice%d_%s",(int)((addr-0x1f801c00)/16),sregname[(addr&15)/2]);
	} else  sprintf(buf,"%08x",(int)addr);
	return buf;
}

#define	spu_reg0	(*(UINT16*)&hwarea[0x1daa])
#define	t0_count	(*(UINT16*)&hwarea[0x1100])
#define	t0_mode		(*(UINT16*)&hwarea[0x1104])
#define	t0_target	(*(UINT16*)&hwarea[0x1108])
#define	t1_count	(*(UINT16*)&hwarea[0x1110])
#define	t1_mode		(*(UINT16*)&hwarea[0x1114])
#define	t1_target	(*(UINT16*)&hwarea[0x1118])
#define	t2_count	(*(UINT16*)&hwarea[0x1120])
#define	t2_mode		(*(UINT16*)&hwarea[0x1124])
#define	t2_target	(*(UINT16*)&hwarea[0x1128])
#define	int_reg		(*(UINT16*)&hwarea[0x1070])
#define	int_mask	(*(UINT16*)&hwarea[0x1074])
#define	dma_pcr		(*(UINT32*)&hwarea[0x10f0])
#define	dma_icr		(*(UINT32*)&hwarea[0x10f4])

static unsigned char hwarea[0x8000];

/*
	counter mode

	base = 0x48
	0x10	肱݋?
	0x01	Q[gL
	0x100	0=clock/8 1=clock
*/

#define	INTERRUPT(no)	reg |= (1<<no)

static UINT32 base_count;
static UINT32 prev_hcount;
static UINT32 t0_limit=0xFFFF;
static UINT32 t1_limit=0xFFFF;
static UINT32 t2_limit=0xFFFF;

int Irq_Pulse = 0;

INT32 Event_Register = 0;
UINT32 Event_List[16];
UINT32 Event_Mask[16];
UINT32 Event_RootCount=0;

#if !AUTOSPEED

#define	PER	64

/* 1Wv̏ߐ */ 
#else
UINT32 cpuspeed,oldcnt;

UINT32 rdtsc()
{
 UINT32 ret;

 asm (" rdtsc "
    : "=a"(ret)
    :
    : "eax", "edx" );

 return ret;
}
#endif

static int treg=0;

int update_hw(void)
{
        int reg = 0,int_happen=0;
#if AUTOSPEED
	UINT32 tmp=rdtsc();
        UINT32 PER=(tmp-oldcnt); // /cpuspeed;

        oldcnt = tmp;
#else
        int tmp;
#endif

	base_count+=PER;

#if AUTOSPEED
	if (base_count>=cpuspeed) {
#else
        if (base_count>=550000) {
#endif
		base_count=0;
		prev_hcount=0;
		PRINTF("VSync\n");
		win_update();
		gpu_update();
		INTERRUPT(INT_VSync);
                Event_RootCount++;
	}

	tmp = t0_count + ((t0_mode & 0x100)?PER:PER/8);
	if (tmp<t0_limit) t0_count=tmp;
	else {
		t0_count = 0;
		if (t0_mode & 0x10) INTERRUPT(INT_CNT0);
	}

	while (base_count>=prev_hcount+2200) {
		prev_hcount+=2200;
		tmp = t1_count + 1;
		if (tmp<t1_limit) t1_count=tmp;
		else {
			t1_count = 0;
			if (t1_mode & 0x10) INTERRUPT(INT_CNT1);
		}
	}

	tmp = t2_count + ((t2_mode & 0x100)?PER:PER/8);
	if (tmp<t2_limit) t2_count=tmp;
	else {
		t2_count = 0;
		if (t1_mode & 0x10) INTERRUPT(INT_CNT1);
	}

	if (dma_icr&0x7f000000) { dma_icr&=0xffffff; INTERRUPT(INT_DMA); }

        if (Event_Register < 0)
        {
            int x,y;

            tmp = Event_RootCount;

            for(x=0;x<16;x++)
            {
                y = Event_Mask[x];
                if (Event_Register & y)
                    if (tmp >= Event_List[x])
                    {
                        treg |= y;
                        Event_Register &= ~y;
                    }
            }
            if (!(Event_Register & 0x7FFFFFFF))
                Event_Register = 0;
        }

        reg |= Irq_Pulse;
        HW32(0x1070) |= reg;
        Irq_Pulse = 0;
        if ((HW32(0x1074) & reg))
        {
            int_happen = 1;
        }

        HW32(0x1070) |= treg;
        treg=0;
        if ((HW32(0x1074) & treg))
        {
            int_happen = 1;
        }
        
    return int_happen;
}

#if 1
int update_counter(void)
{
	return update_hw();
}

#else
int update_counter(void)
{
	int reg = 0,int_happen=0;

	base_count++;
	if (base_count>=550000/2) { /* 33Mhz x 1/60 */
		base_count=0;
		PRINTF("VSync\n");
		win_update();
		gpu_update();
		INTERRUPT(INT_VSync);
                Event_RootCount++;
	}
	if ((t0_mode & 0x100) || ((base_count&7)==0)) {
		if (++t0_count>=t0_target) {
			t0_count = 0;
			if (t0_mode & 0x10) INTERRUPT(INT_CNT0);
		}
	}
	if ((t1_mode & 0x100) || ((base_count&7)==0)) {
		if (++t1_count>=t1_target) {
			t1_count = 0;
			if (t1_mode & 0x10) INTERRUPT(INT_CNT1);
		}
	}
	if ((t2_mode & 0x100) || ((base_count&7)==0)) {
		if (++t2_count>=t2_target) {
			t2_count = 0;
			if (t2_mode & 0x10) INTERRUPT(INT_CNT2);
		}
	}
	if (dma_icr&0x7f000000) { dma_icr&=0xffffff; INTERRUPT(INT_DMA); }

        reg |= Irq_Pulse;
        HW32(0x1070) |= reg;
        Irq_Pulse = 0;
        if ((HW32(0x1074) & reg))
        {
            int_happen = 1;
        }

        HW32(0x1070) |= treg;
        treg=0;
        if ((HW32(0x1074) & treg))
        {
            int_happen = 1;
        }
        
    return int_happen;
}
#endif

/* OT clear */
void dma6_exec(UINT32 adr,UINT32 bcr,UINT32 chcr)
{
	if (chcr!=0x11000002) {
		PRINTF("dma6 unknown %x\n",(int)chcr);
		return;
	}
	adr&=0xffffff;
	while(bcr--){
		*(UINT32*)&ram[adr] = (adr-4)&0xffffff;
		adr-=4;
	}
}

int hw_init(void)
{
 int x;
#if AUTOSPEED
 time_t t1;
 UINT32 clk1 = rdtsc();

 t1 = clock() + CLOCKS_PER_SEC;
 while (clock() < t1);
 oldcnt = rdtsc();
 cpuspeed = (oldcnt - clk1);
 printf("CPU speed: %d MHz\n",(int)(cpuspeed/1000000));
 cpuspeed /= (60); //*550000);
#endif

 for (x=0;x<16;x++)
    Event_Mask[x] = 1<<x;

 DrawInit();
 mdec_init();
 sio_init();
 cd_init();
 win_init();
 spu_init();
 vram = win_createbitmap(1024,512,16);

 return FPSE_OK;
}

void hw_close(void)
{
 cd_term();
 spu_term();
}

UINT8 hw_read8(UINT32 adr)
{
	int ret;
	switch(adr){
	case 0x1f801040: ret = sio_readdata8(&Sio0); break;
	case 0x1f801050: ret = sio_readdata8(&Sio1); break;
	case 0x1f801800: ret = cd0_read(); break;
	case 0x1f801801: ret = cd1_read(); break;
	case 0x1f801802: ret = cd2_read(); break;
	case 0x1f801803: ret = cd3_read(); break;
	default:
		ret = hwarea[adr-0x1f800000]; break;
	}
#if NOPRINTPC
	printpc();
#endif
	PRINTF("read byte %s,%02x\n",hwname(adr),ret);
	return ret;
}

UINT16 hw_read16(UINT32 adr)
{
	int ret;

	switch(adr){
	case 0x1f801070:
        if (Event_Register < 0)
        {
            int x,y;

            ret = Event_RootCount;
            Event_RootCount++;
            for(x=0;x<16;x++)
            {
                y = Event_Mask[x];
                if (Event_Register & y)
                    if (ret >= Event_List[x])
                    {
                        treg |= y;
                        Event_Register &= ~y;
                    }
                if (!(Event_Register & 0x7FFFFFFF))
                    Event_Register = 0;
            }
        }
        ret= HW16(0x1070)|treg; break;

/* SIO0 - Controllers & MemCards */
	case 0x1f801040: ret = sio_readdata16(&Sio0); break;
	case 0x1f801044: ret = Sio0.Status; break;
	case 0x1f80104a: ret = Sio0.Ctrl.Control16; break;
	case 0x1f80104e: ret = Sio0.Baud; break;
/* SIO1 - Serial port */
	case 0x1f801050: ret = sio_readdata16(&Sio1); break;
	case 0x1f801054: ret = Sio1.Status; break;
	case 0x1f80105a: ret = Sio1.Ctrl.Control16; break;
	case 0x1f80105e: ret = Sio1.Baud; break;

	default:
		if (adr>=0x1f801c00 && adr<0x1f801e00)
			ret = spu_read(adr);
                else ret = HW16(adr & 0xFFFF);
        }
#if NOPRINTPC
	printpc();
#endif
	PRINTF("read short %s,%04x\n",hwname(adr),ret);
	return ret;
}

UINT32 hw_read32(UINT32 adr)
{
	UINT32 ret;

	switch(adr) {
	case 0x1f801810: ret = gp0_read(); break;
	case 0x1f801814: ret = gp1_read(); break;
/* SIO0 */
	case 0x1f801040: ret = sio_readdata32(&Sio0); break;
/* SIO1 */
	case 0x1f801050: ret = sio_readdata32(&Sio1); break;

	case 0x1f801070:
        if (Event_Register < 0)
        {
            int x,y;

            ret = Event_RootCount;
            Event_RootCount++;
            for(x=0;x<16;x++)
            {
                y = Event_Mask[x];
                if (Event_Register & y)
                    if (ret >= Event_List[x])
                    {
                        treg |= y;
                        Event_Register &= ~y;
                    }
                if (!(Event_Register & 0x7FFFFFFF))
                    Event_Register = 0;
            }
        }
        ret= HW32(0x1070)|treg; break;

	case 0x1f801088:
	case 0x1f801098:
	case 0x1f8010a8:
	case 0x1f8010b8:
	case 0x1f8010c8:
	case 0x1f8010d8:
	case 0x1f8010e8:
		/* dma_chcr */
		/* busyłȂȂ read 0x01000000 ς ? */
		ret = 0;
		break;
	case 0x1f8010f4:
		/* dma status */
		ret = 0;
		break;
	default:
		ret = *(UINT32*)&hwarea[adr-0x1f800000];
		break;
	}
#if NOPRINTPC
	printpc();
#endif
	PRINTF("read long %s,%08x\n",hwname(adr),(int)ret);
	return ret;
}

void hw_write8(UINT32 adr,UINT32 data)
{
#if NOPRINTPC
	printpc();
#endif
	PRINTF("write byte %s,%02x\n",hwname(adr),(int)data);

	switch(adr){
	case 0x1f801040: sio_writedata8(&Sio0,data); break;
	case 0x1f801050: sio_writedata8(&Sio1,data); break;
	case 0x1f801800: cd0_write(data); break;
	case 0x1f801801: cd1_write(data); break;
	case 0x1f801802: cd2_write(data); break;
	case 0x1f801803: cd3_write(data); break;
	default:
		hwarea[adr-0x1f800000]=data;
		break;
	}
}

void hw_write16(UINT32 adr,UINT32 data)
{
	switch(adr) {
	case 0x1f801040: sio_writedata16(&Sio0,data); break;
	case 0x1f801048: sio_mode_write(&Sio0,data); break;
	case 0x1f80104a: sio_control_write(&Sio0,data); break;
	case 0x1f80104e: sio_baud_write(&Sio0,data); break;

	case 0x1f801050: sio_writedata16(&Sio1,data); break;
	case 0x1f801058: sio_mode_write(&Sio1,data); break;
	case 0x1f80105a: sio_control_write(&Sio1,data); break;
	case 0x1f80105e: sio_baud_write(&Sio1,data); break;

	case 0x1f801070:
		/* interrupt clear? */
                HW16(0x1070) &= (data & HW16(0x1074));
                treg &= (data & HW16(0x1074));
		break;
	default:
		if (adr>=0x1f801c00 && adr<0x1f801e00) {
			spu_write(adr,data);
			break;
		}
		*(UINT16*)&hwarea[adr-0x1f800000]=data;
		break;
	}
#if NOPRINTPC
	printpc();
#endif
	PRINTF("write short %s,%04x\n",hwname(adr),(int)data);
}

void hw_write32(UINT32 adr,UINT32 data)
{
	switch(adr) {
	case 0x1f801040: sio_writedata32(&Sio0,data); break;
	case 0x1f801050: sio_writedata32(&Sio1,data); break;

	case 0x1f801070:
		/* XgAĂ interrupt clear? */
                HW32(0x1070) &= (data & HW32(0x1074));
                treg &= (data & HW32(0x1074));
		break;
	case 0x1f801810: gp0_write(data); return;
	case 0x1f801814: gp1_write(data); return;

#define	DMA_ENABLE(n)	(dma_pcr&(8<<(n*4)))
#define	DMA_INTERRUPT(n)	if (dma_icr&(1<<(16+n))) dma_icr |= (1<<(24+n));

	case 0x1f801088: /* MDECin */
		if (DMA_ENABLE(0)) {
			dma0_exec(HW32(0x1080),HW32(0x1084),data);
			HW32(0x1088) = data&~0x01000000;
			DMA_INTERRUPT(0);
		}
		break;
	case 0x1f801098: /* MDECout */
		if (DMA_ENABLE(1)) {
			dma1_exec(HW32(0x1090),HW32(0x1094),data);
			HW32(0x1098) = data&~0x01000000;
			DMA_INTERRUPT(1);
		}
		break;
	case 0x1f8010a8: /* GPU DMA */
		if (DMA_ENABLE(2)) {
			gpudma_exec(HW32(0x10a0),HW32(0x10a4),data);
			HW32(0x10A8) = data&~0x01000000;
			DMA_INTERRUPT(2);
		}
		break;
	case 0x1f8010b8: /* CDROM DMA */
		if (DMA_ENABLE(3)) {
			dma3_exec(HW32(0x10b0),HW32(0x10b4),data);
			HW32(0x10B8) = data&~0x01000000;
			DMA_INTERRUPT(3);
		}
		break;
	case 0x1f8010c8: /* SPU DMA */
		if (DMA_ENABLE(4)) {
			dma4_exec(HW32(0x10c0),HW32(0x10c4),data);
			HW32(0x10C8) = data&~0x01000000;
			DMA_INTERRUPT(4);
		}
		break;
#if 0
	case 0x1f8010d8: /* PIO DMA */
		if (DMA_ENABLE(5)) {
			dma5_exec(HW32(0x10D0),HW32(0x10D4),data);
			HW32(0x10D8) = data&~0x01000000;
			DMA_INTERRUPT(5);
		}
		break;
#endif
	case 0x1f8010e8: /* OT clear DMA */
		if (DMA_ENABLE(6)) {
			dma6_exec(HW32(0x10E0),HW32(0x10E4),data);
			HW32(0x10E8) = data&~0x01000000;
			DMA_INTERRUPT(6);
		}
		break;
	case 0x1f8010f4:
		HW32(0x10f4)=data&0xffffff;
		break;
        case 0x1F801104:
                HW32(0x1104) = data;
                if (data & 0x08) t0_limit = t0_target;
                            else t0_limit = 0xFFFF;
                break;
        case 0x1F801114:
                HW32(0x1114) = data;
                if (data & 0x08) t1_limit = t1_target;
                            else t1_limit = 0xFFFF;
                break;
        case 0x1F801124:
                HW32(0x1124) = data;
                if (data & 0x08) t2_limit = t2_target;
                            else t2_limit = 0xFFFF;
                break;
	case 0x1f801820: mdec0_write(data); break;
	case 0x1f801824: mdec1_write(data); break;

	default:
		*(UINT32*)&hwarea[adr-0x1f800000]=data;

	}
#if NOPRINTPC
	printpc();
#endif
	PRINTF("write long %s,%08x\n",hwname(adr),(int)data);
}

/*
	dma_icr(1f8010f4)
	read: 0x7f000000 .. 肱ݔ	1=
	write:0x7f       .. 1bit0ɂ?
	r/w       7f     .. 肱݃}XN	1=ok 0=mask
	w         800000 .. ݎset?
*/
