#include <windows.h>
#include <io.h>
#include <memory.h>
#include <fcntl.h>

#include "fpse.h"

#define ACKPAD   sio->Status |= DSR;
#define NACKPAD  sio->Status &= ~DSR;

#define B13         0x2000
#define READ_CARD   0x80

#define KPAD    1
#define KMEM    2

// This struct should be allocate in SIO_Type in the future...
typedef struct {
    int  type;
    int  ptr;
    char buf[16];
/* Memory card fields */
    char membuf[256];
    union {
        UINT8  l[4];
        UINT32 hl;
    } addr;
    int  endflg,crc;
    char *cardname;
    char memorycard[128*1024];
} pad_type;

static pad_type *flx=NULL;
static pad_type  pad[2];
static int       nflx=0;
static char      cd1[] = {"slot1"};
static char      cd2[] = {"slot2"};

int PAD_Init(SIO_Type *sio)
{
 int f,x;

 memset(pad,0,sizeof(pad));
 pad[0].cardname = cd1;
 pad[1].cardname = cd2;

 for (x=0;x<2;x++)
    {
// GamePad header
     pad[x].buf[1] = 0x41;
     pad[x].buf[2] = 0x5A;

// Memory card header
     pad[x].membuf[2] = 0x5A;
     pad[x].membuf[3] = 0x5D;

// Load memory card
     f = open(pad[x].cardname, O_RDONLY|O_BINARY);
     if (f!=-1)
        {
            read(f,pad[x].memorycard,128*1024);
            close(f);
        }
    }

// Times to wait before irq
 sio->Flags = 0x80000003;

// Init externals
 JOY_Open();

 return FPSE_OK;
}

void PAD_Close(SIO_Type *sio)
{
}

void PAD_WriteData(SIO_Type *sio)
{
 UINT8 *buf = sio->TxBuf;
 int    v;
 UINT8  crc;

 if (flx == NULL) return;

 if (flx->ptr == 0)
    {
        switch (*buf) {
        case 0x01: flx->type = KPAD; break;
        case 0x81: flx->type = KMEM; break;
        default: return;
        }
        flx->ptr = 1;
        ACKPAD
        sio_async(sio,flx->buf,1);
        NACKPAD
        return;
    }

 if (flx->type == KPAD)
    {
        if (flx->ptr == 1)
            {
                if (*buf != 0x42)
                    {
                        flx->ptr = 0;
                        return;
                    }
                else 
                    {
                        v = JOY_Poll(nflx);
                        flx->buf[3] = (v >> 8); // | 0x06;
                        flx->buf[4] = v & 0xFF;
                    }
            } else JOY_Vibrate(*buf);
        if (flx->ptr < 4) ACKPAD
        sio_async(sio,flx->buf + flx->ptr,1);
        NACKPAD
        if (++flx->ptr >= 5) flx->ptr = 0;
        return;
    }

 if ((flx->type & 0x7F) == KMEM)
    {
        if (flx->ptr == 1)
            {
                switch (*buf) {
                case 'R':
                    flx->type |= READ_CARD;
                    flx->endflg = 139;
                    break;
                case 'W':
                    flx->endflg = 138;
                    break;
                default:
                    flx->ptr = 0;
                    return;
                }
            }
        if (flx->type & READ_CARD)
            {
                switch (flx->ptr) {
                case 4:
                    flx->addr.l[1] = *buf;
                    flx->membuf[5] = *buf;
                    flx->membuf[6] = 0x5C;
                    flx->membuf[7] = 0x5D;
                    flx->membuf[8] = *buf;
                    break;
                case 5:
                    flx->addr.l[0] = *buf;
                    flx->membuf[9] = *buf;
                    memcpy(flx->membuf+10,
                           flx->memorycard+flx->addr.hl*128,
                           128);
                    crc = 0; // pad[flx].addr.l[0] ^ pad[flx].addr.l[1];
                    for (v=8;v<128+10;v++)
                        crc ^= flx->membuf[v];
                    flx->membuf[v] = crc;
                    flx->membuf[v+1] = 0x47;
                    break;
                }
            }
        else {
                switch (flx->ptr) {
                case 4:
                    flx->addr.l[1] = *buf;
                    flx->memorycard[135] = 0x5C;
                    flx->memorycard[136] = 0x5D;
                    flx->memorycard[137] = 0x47;
                    break;
                case 5:
                    flx->addr.l[0] = *buf;
                    flx->crc = flx->addr.l[0] ^ flx->addr.l[1];
                    break;
                case 134:
                    if (*buf != flx->crc)
                    {
                        flx->memorycard[137] = 0x4E;
                        break;
                    }
                    memcpy(flx->memorycard+flx->addr.hl*128,
                           flx->membuf+7,128);
                }
                if (flx->ptr > 3 &&
                    flx->ptr < 134)
                {
                    flx->membuf[flx->ptr] = *buf;
                    flx->crc ^= *buf;
                }
            }
        if (flx->ptr < flx->endflg) ACKPAD
        sio_async(sio,flx->membuf + flx->ptr,1);
        NACKPAD
        if (++flx->ptr >= flx->endflg) flx->ptr = 0;
//        pad[flx].ptr = 0;
//        return;
    }
}

void PAD_UpdateStatus(SIO_Type *sio)
{
 if (!(sio->Ctrl.Control16 & 2)) {
    pad[0].ptr = pad[1].ptr = 0;
    flx = NULL;
    return;
 }

 if (sio->Ctrl.Control16 & B13) nflx = 1;
                           else nflx = 0;
 flx = pad + nflx;
}
