/*
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "spu.h"
#include "xa.h"
#include "sound.h"
#include "fpse.h"

//#define	PRINTF if (verbose&VERBOSE_SPU) printf

#define	RAMADDR(addr) ((void*)&ram[(addr)&0xffffff])
#define	SPU_DATA(adr)	spureg[(adr-0x1f801c00)/2]

static int spu_sbaddr;
static unsigned short spureg[(0x1e00-0x1c00)/2];
static unsigned char spumem_flag[0x10000/8];
static short *wavebuf;

typedef struct {
	short vol_l,vol_r;
	short pitch;
	short addr;
	short adsr1,adsr2;
	short adsrvol;
	short repeat_addr;
} voice_data;

#define	WCACHE_MAX	32

typedef struct {
	int addr;
	int size,loop_start;
	short *buf;
} WAVE_CACHE;

static WAVE_CACHE wcache[WCACHE_MAX];
static int wcache_index;


/* bit flag treat */

#define	setbit(buf,offset)   \
	(buf)[(offset)/(sizeof(buf[0])*8)]|=1<<((offset)&(sizeof(buf[0])*8-1))
#define	resetbit(buf,offset) \
	(buf)[(offset)/(sizeof(buf[0])*8)]&=~(1<<((offset)&(sizeof(buf[0])*8-1)))
#define	getbit(buf,offset)  \
	((buf)[(offset)/(sizeof(buf[0])*8)]&(1<<((offset)&(sizeof(buf[0])*8-1))) )

static void setbits(unsigned char *buf,int offset,int size)
{
	unsigned char lmask,rmask;
	int size2,end;

	end = offset+size;
	lmask =   0xff<<(offset&7);
	rmask = ~(0xff<<(end&7));

	size2 = end/8 - offset/8;
	if (size2==0) {
		buf[offset/8]|= lmask&rmask;
	} else {
		buf[offset/8]|=lmask;
		buf[end/8]|=rmask;
		if (size2>1) memset(&buf[offset/8+1],0xff,size2-1);
	}
}



/*
static void dump(unsigned char *p,int len)
{
	int i;
	for(i=0;i<len;i++)
		printf("%02x ",*p++);
	printf("\n");
}
*/

static void vag2wave(WAVE_CACHE *wc,int addr)
{
	unsigned char *vag;
	short *wav;
	int n,loop_start,pos;
	XA_PREV prev;

	vag = spumem + addr*8;
	wav = wavebuf + addr*28;

	decode_xa_init(&prev);
	n = 0;
	loop_start = -1;
	pos = addr;
/*
	flags
	0x00: normal samples
	0x02: repeat samples
	0x03: end of samples, start from repeat start
	0x06: repeat start
	0x07: end of samples
*/

	while(1) {
		int flags = vag[1];
	//	dump(vag,16);
		decode_vag(wav+n,vag,&prev);
		if (flags==0x06) loop_start = n;
		vag+=16;
		n+=28;
		if (flags==0x03 || flags==0x07) break;
		/* for err data */
		pos+=2;
		if (pos>=0x10000) {
			n=28;
			break;
		}
	}

	printf("wav:%x %d %d\n",addr,n,loop_start);

	wc->addr = addr;
	wc->buf = wav;
	wc->size = n;
	wc->loop_start = loop_start;
}

WAVE_CACHE *get_wcache(int addr)
{
	WAVE_CACHE *wc;

	if (getbit(spumem_flag,addr)) {
		resetbit(spumem_flag,addr);
	} else {
		int i;

		wc = wcache;
		for(i=0;i<WCACHE_MAX;i++,wc++) {
			if (addr == wc->addr) return wc;
		}
	}

	wc = &wcache[wcache_index];
	wcache_index = (wcache_index+1)&(WCACHE_MAX-1);
	vag2wave(wc,addr);

	return wc;
}

static void KeyOn(int ch)
{
	voice_data *v = (voice_data*)&spureg[ch*8];
	int freq = v->pitch*44100/4096;
	WAVE_CACHE *wc = get_wcache(v->addr);

	if (wc->loop_start>=0) v->adsrvol=0xff;

	wav_play(ch,wc->buf,wc->size,freq,v->vol_l,v->vol_r,wc->loop_start);
}

static void KeyOff(int ch)
{
	voice_data *v = (voice_data*)&spureg[ch*8];
	v->adsrvol = 0;
	wav_stop(ch);
}

int spu_init(void)
{
	wavebuf = malloc(0x8000*28*sizeof(short));
	return sound_init();
}

int spu_term(void)
{
	free(wavebuf);
	return sound_term();
}

extern char* hwname(unsigned long addr);

int spu_read(UINT32 adr)
{
	int ret;
	switch(adr){
	case 0x1f801d9c:
		{ int i;
		ret = 0;
		for(i=0;i<8;i++)
			ret |= wav_status(i)<<i;
		}
		break;
	case 0x1f801d9e:
		{ int i;
		ret = 0;
		for(i=0;i<4;i++)
			ret |= wav_status(i+4)<<i;
		}
		break;
	default:
		ret = SPU_DATA(adr);
	}
	PRINTF("spu_read:%s,%04x\n",hwname(adr),ret);
	return ret;
}

void spu_write(UINT32 adr,int data)
{
	if (adr!=0x1f801da8)
		PRINTF("spu_write:%s,%04x\n",hwname(adr),data);
	SPU_DATA(adr) = data;
	switch(adr) {
	case 0x1f801d80:
	case 0x1f801d82:
		SetMainVolume(SPU_DATA(0x1f801d80),SPU_DATA(0x1f801d82));
		break;
	case 0x1f801d88:
		if (data) {
		int i;
		for(i=0;i<8;i++,data>>=1)
			if (data&1) KeyOn(i);
		}
		break;
	case 0x1f801d8a:
		if (data) {
		int i;
		for(i=0;i<4;i++,data>>=1)
			if (data&1) KeyOn(i+8);
		}
		break;
	case 0x1f801d8c:
		if (data) {
		int i;
		for(i=0;i<8;i++,data>>=1)
			if (data&1) KeyOff(i);
		}
		break;
	case 0x1f801d8e:
		if (data) {
		int i;
		for(i=0;i<4;i++,data>>=1)
			if (data&1) KeyOff(i+8);
		}
		break;
	case 0x1f801da6: /* spu sbaddr */
		spu_sbaddr = data*8;
		break;
	case 0x1f801da8: /* spu data */
	//	if ((spu_reg0&0x30)==0x10) {
			/* non dma mode */
			spumem[spu_sbaddr++]=data;
			spumem[spu_sbaddr++]=data>>8;
			setbit(spumem_flag,spu_sbaddr/8);
			return;
	//	}
		break;
	default:
/*
		if (adr<=0x1f801d00) {
			printf("spu_write:%s,%04x\n",hwname(adr),data);
		}
*/
		break;
	}
}

void dma4_exec(UINT32 adr,UINT32 bcr,UINT32 chcr)
{
	int size;

	size = (bcr>>16)*(bcr&0xffff);

	switch(chcr) {
	case 0x01000201:
#if 0
		{int i;
		for(i=0;i<size*4;i+=16) {
			printf("%04x:",(spu_sbaddr/8)+i/16);
			dump(RAMADDR(adr)+i,16);
		}
		}
#endif
		memcpy(spumem+spu_sbaddr,RAMADDR(adr),size*4);
		setbits(spumem_flag,spu_sbaddr/8,size/8);
		break;
	case 0x01000200:
		memcpy(RAMADDR(adr),spumem+spu_sbaddr,size*4);
		break;
	}
	PRINTF("spu_dma:%08lx %08lx %08lx\n",adr,bcr,chcr);
}
