#include <windows.h>
#include <stdio.h>
#include <memory.h>
#include <malloc.h>

#include "fpse.h"

#define FRAMESKIP       1

#define DIB_RGB_COLORS  0

#define	USEDIBSECTION	1
#define	USE_D3D	        0
#define	USE_GL          0

typedef struct {
	BITMAPINFO bi;
	LPSTR lpBits;
	HBITMAP hbm;
} DIB;

#if FRAMESKIP
int framecount=0;
int frameskip=3;
#endif

void SetDIBHeader(BITMAPINFO *lpbi,int width,int height,int bits)
{
	BITMAPINFOHEADER *lpbih = &lpbi->bmiHeader;
	lpbih->biSize   = sizeof(BITMAPINFOHEADER);
	lpbih->biWidth  = width;
	lpbih->biHeight = -height;
	lpbih->biPlanes = 1;
	lpbih->biBitCount = bits;
	lpbih->biCompression = BI_RGB;
//	lpbih->biCompression = BI_BITFIELDS;
	lpbih->biSizeImage = (((((width*bits)+7)/8)+3)&~3)*height;
	lpbih->biXPelsPerMeter = 0;
	lpbih->biYPelsPerMeter = 0;
	lpbih->biClrUsed = 0;
	lpbih->biClrImportant = 0;
/*	don't work win95
	lpbi->dwMask[2] = 0x7c00;
	lpbi->dwMask[1] = 0x03e0;
	lpbi->dwMask[0] = 0x001f;
*/
}


void StretchDIB(HDC hdc,DIB *dib,int x,int y,int w,int h)
{
#if USEDIBSECTION
	HDC memDC;
	HBITMAP hbmOld;
//	int width  = dib->bi.bmiHeader.biWidth;
//	int height = dib->bi.bmiHeader.biHeight;
//	if (height<0) height= -height;

	memDC = CreateCompatibleDC(hdc);
	hbmOld = SelectObject(memDC,dib->hbm);

	BitBlt(hdc,0,0,w,h,memDC,x,y,SRCCOPY);

	SelectObject(memDC,hbmOld);
	DeleteDC(memDC);
#else
//	int width  = dib->bi.bmiHeader.biWidth;
//	int height = dib->bi.bmiHeader.biHeight;
//	if (height<0) height= -height;

	StretchDIBits(hdc,
		0,0,w,h, 
		x,y,w,h, 
	dib->lpBits,&dib->bi,DIB_RGB_COLORS,SRCCOPY);
/*
	SetDIBitsToDevice(hdc,
		0,0,w,h, 
		x,y,x,h,
	dib->lpBits,&dib->bi,DIB_RGB_COLORS);
*/
#endif
}

	static DIB dib;
	static HWND hwnd;
	static int ofs_x,ofs_y,disp_w,disp_h;

LRESULT APIENTRY WndFunc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
	HDC	hdc;

	switch (msg) {
	case WM_KEYDOWN:
		switch(wParam) {
		case VK_ESCAPE:
			DestroyWindow(hwnd);
			break;
                case VK_F12:
                case VK_F10:
                case VK_F1:
                case 'S':
                        ScreenShot();
                        break;
		}
		break;
#if USE_D3D
	case WM_MOVE:
	case WM_SIZE:
		SetScreenRect(LOWORD(lParam),HIWORD(lParam));
		break;
#endif
	case WM_PAINT:
#if	USE_D3D
//		D3D_update();
		return DefWindowProc(hwnd,msg,wParam,lParam);
#elif USE_GL
		return DefWindowProc(hwnd,msg,wParam,lParam);
#else
		if (dib.lpBits) {
			PAINTSTRUCT ps;
			static int t0,t1,frame,fps;

			hdc = BeginPaint(hwnd,&ps);
#if FRAMESKIP
                        if (++framecount > frameskip)
			    StretchDIB(hdc,&dib,ofs_x,ofs_y,disp_w,disp_h);
#endif
			if ((t1=GetTickCount()-t0)>1000) {
				fps = 1000*frame/t1;
				t0+=t1;
				frame=0;
			}
			frame++;
#if FRAMESKIP
                        if (framecount > frameskip)
#endif
			{
				char szbuf[20];
				wsprintf(szbuf,"fps:%d",fps);
				TextOut(hdc,0,0,szbuf,strlen(szbuf));
#if FRAMESKIP
                                framecount = 0;
#endif
			}
			EndPaint(hwnd,&ps);
		}
#endif
		break;
	case WM_DESTROY:
		hw_close();
#if USE_D3D
		D3D_exit();
#endif

#if USEDIBSECTION
		DeleteObject(dib.hbm);
#else
		free(dib.lpBits);
#endif
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hwnd,msg,wParam,lParam);
	}
	return 0;
}

void *win_createbitmap(int w,int h,int bits)
{
#if USEDIBSECTION
	HDC hdc;
#endif
	SetDIBHeader(&dib.bi,w,h,bits);
#if USEDIBSECTION
	hdc = GetDC(hwnd);
	dib.hbm = CreateDIBSection(hdc,(BITMAPINFO*)&dib.bi,DIB_RGB_COLORS,(void**)&dib.lpBits,NULL,0);
	if (dib.hbm==NULL) MessageBox(hwnd,"createdib false",NULL,MB_OK);
	ReleaseDC(hwnd,hdc);

	memset(dib.lpBits,0,dib.bi.bmiHeader.biSizeImage);
#else
	dib.lpBits = malloc(dib.bi.bmiHeader.biSizeImage);
#endif
        memset((UINT16 *)dib.lpBits,0,1024*512*2);

	return dib.lpBits;
}

/* display offset */
void win_offset(int x,int y)
{
	ofs_x = x;
	ofs_y = y;
#if	USE_D3D
	D3D_offset(x,y);
#endif
}

#if 0
static void print_rect(RECT *rect)
{
	printf("%d %d %d %d\n",rect->left,rect->top,rect->right-rect->left,rect->bottom-rect->top);
}
#endif

void win_resize(int w,int h)
{
	RECT window,client;
	int ew,eh;

	GetWindowRect(hwnd,&window);
	GetClientRect(hwnd,&client);
//	printf("window");print_rect(&window);
//	printf("client");print_rect(&client);
	ew = (window.right-window.left)-(client.right-client.left),
	eh = (window.bottom-window.top)-(client.bottom-client.top),

	disp_w = w; 
	disp_h = h; 
//	printf("%d %d %d %d %d %d\n",w,h,w+ew,h+eh,ew,eh);
	MoveWindow(hwnd,window.left,window.top,w+ew,h+eh,TRUE);
}

void win_init(void)
{
	static char szAppName[] = "FPSE";
	static char szClassName[] = "FPSE";
	WNDCLASS wndclass;

	HINSTANCE hInst = 0;

	wndclass.style = 0; //CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndFunc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInst;
	wndclass.hIcon = LoadIcon(NULL,IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL,IDC_ARROW);
	wndclass.hbrBackground = GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szClassName;

	RegisterClass(&wndclass);

	hwnd = CreateWindow(
		szClassName, //classname
		szAppName, //title
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT,
		CW_USEDEFAULT,
		320,
		240,
		NULL, //parent
		NULL, //menu
		hInst,
		NULL);

	ShowWindow(hwnd,SW_SHOW);
	UpdateWindow(hwnd);
	win_resize(320,240);

#if	USE_D3D
	D3D_init(hwnd);
#elif USE_GL
	gl_init(hwnd);
#endif
}

void win_update(void)
{
	MSG msg;
/*
	HDC hdc = GetDC(NULL);
	StretchDIB(hdc,&dib,0,0,1024,512);
	ReleaseDC(NULL,hdc);
*/
#if USE_D3D
	D3D_update();
#else
	InvalidateRect(hwnd,NULL,FALSE);
#endif
	while (PeekMessage(&msg,NULL,0,0,PM_REMOVE)) {
		if (msg.message==WM_QUIT) {
			exit(msg.wParam);
		}
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}
