/*
	freq = 37800
*/
#include "xa.h"

static void decode_xa(short *out,unsigned char *in,XA_PREV *prev);
static void decode_xa2(short *out,unsigned char *in,XA_PREV *prev);

void decode_xa_init(XA_PREV *prev)
{
	prev->s1 = 0;
	prev->s2 = 0;
}

void decode_xa_mono(short *out,unsigned char *in)
{
	int i;
	XA_PREV prev[1];

	decode_xa_init(prev);
	for(i=0;i<18;i++) {
		decode_xa(out,in,prev);
		in+=128;
		out+=28*8;
	}
}

void decode_xa_stereo(short *out,unsigned char *in)
{
	int i;
	XA_PREV prev[2];

	decode_xa_init(prev);
	decode_xa_init(prev+1);
	for(i=0;i<18;i++) {
		decode_xa2(out,in,prev);
		in+=128;
		out+=28*8;
	}
}


const static int f[5][2] = {
	{0,0},
	{60,0},
	{115,-52},
	{98,-55},
	{122,-60}
};

static void decode_xa(short *out,unsigned char *in,XA_PREV *prev)
{

	int i,j;
	int shift,filter,f0,f1;
	int s_1,s_2;

	
	s_1 = prev->s1;
	s_2 = prev->s2;

  for(i=0;i<4;i++) {
	shift  = in[4+i*2] & 15;
	filter = in[4+i*2] >> 4;
	f0 = f[filter][0];
	f1 = f[filter][1];

	for(j=0;j<28;j++) {
		int d,s,t;
		d = in[16+i+j*4];

		t = d&0xf;
		if (t>=8) t-=16;
		s = ( t<<(12-shift) ) + ((s_1*f0 + s_2*f1+32)>>6);
		if (s<-32768) s=-32768; else if (s>32767) s=32767;
		*out++ = s;
		s_2 = s_1;
		s_1 = s;
	}

	shift  = in[5+i*2] & 15;
	filter = in[5+i*2] >> 4;
	f0 = f[filter][0];
	f1 = f[filter][1];

	for(j=0;j<28;j++) {
		int d,s,t;
		d = in[16+i+j*4];

		t = d>>4;
		if (t>=8) t-=16;
		s = ( t<<(12-shift) ) + ((s_1*f0 + s_2*f1+32)>>6);
		if (s<-32768) s=-32768; else if (s>32767) s=32767;
		*out++ = s;
		s_2 = s_1;
		s_1 = s;
	}
  }

	prev->s1 = s_1;
	prev->s2 = s_2;
}


static void decode_xa2(short *out,unsigned char *in,XA_PREV *prev)
{
	int i,j,filter;
	int shift_0,f0_0,f1_0;
	int shift_1,f0_1,f1_1;

  for(i=0;i<4;i++) {

	shift_0 = in[4+i*2] & 15;
	filter  = in[4+i*2] >> 4;
	f0_0 = f[filter][0];
	f1_0 = f[filter][1];
	
	shift_1 = in[5+i*2] & 15;
	filter  = in[5+i*2] >> 4;
	f0_1 = f[filter][0];
	f1_1 = f[filter][1];
	
	for(j=0;j<28;j++) {
		int d,s,t;

		d = in[16+i+j*4];

		t = d&0xf;
		if (t>=8) t-=16;
		s = ( t<<(12-shift_0) ) + ((prev[0].s1*f0_0 + prev[0].s2*f1_0+32)>>6);
		if (s<-32768) s=-32768; else if (s>32767) s=32767;
		out[0] = s;
		prev[0].s2 = prev[0].s1;
		prev[0].s1 = s;

		t = d>>4;
		if (t>=8) t-=16;
		s = ( t<<(12-shift_1) ) + ((prev[1].s1*f0_1 + prev[1].s2*f1_1+32)>>6);
		if (s<-32768) s=-32768; else if (s>32767) s=32767;
		out[1] = s;
		prev[1].s2 = prev[1].s1;
		prev[1].s1 = s;
		out+=2;
	}
  }
}

void decode_vag(short *out,unsigned char *in,XA_PREV *prev)
{
	int j;
	int shift,filter,f0,f1;
	int s_1,s_2;

	
	s_1 = prev->s1;
	s_2 = prev->s2;

	shift  = in[0] & 15;
	filter = in[0] >> 4;
	f0 = f[filter][0];
	f1 = f[filter][1];

	for(j=0;j<14;j++) {
		int d,s,t;
		d = in[2+j];

		t = d&0xf;
		if (t>=8) t-=16;
		s = ( t<<(12-shift) ) + ((s_1*f0 + s_2*f1+32)>>6);
		if (s<-32768) s=-32768; else if (s>32767) s=32767;
		*out++ = s;
		s_2 = s_1;
		s_1 = s;

		t = d>>4;
		if (t>=8) t-=16;
		s = ( t<<(12-shift) ) + ((s_1*f0 + s_2*f1+32)>>6);
		if (s<-32768) s=-32768; else if (s>32767) s=32767;
		*out++ = s;
		s_2 = s_1;
		s_1 = s;
	}

	prev->s1 = s_1;
	prev->s2 = s_2;
}
